﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace Clock {
	public partial class Clock : Form {

		private Bitmap GlassBlur;
		private Graphics GlassBlurGraphics;
		private Bitmap GlassBlurScratch;
		private Graphics GlassBlurScratchGraphics;
		private Bitmap FinalClockFace;
		private Graphics FinalClockFaceGraphics;

		private Rectangle FinalClockFaceArea;

		private Size CopyFromScreenSize;

		private RectangleF GlassScratchArea;

		private RectangleF ScratchToScreenArea;

		private SolidBrush TintBrush = new SolidBrush(Color.Black);

		

		private int[] AlphaWorkerArea;
		private int[] FaceAlpha;

		public Color TintColor {
			get { return this.TintBrush.Color; }
			set { this.TintBrush.Color = value; }
		}


		private int blurFactor = 4;
		private int BlurFactor {
			get { return this.blurFactor; }
			set { this.blurFactor = value;
				if (this.GlassBlur != null) this.GlassBlur.Dispose(); this.GlassBlur = null;
				if (this.GlassBlurScratch != null) this.GlassBlurScratch.Dispose(); this.GlassBlurScratch = null;
				if (this.FinalClockFace != null) this.FinalClockFace.Dispose(); this.FinalClockFace = null;
			}
		}



		public Clock() {
			InitializeComponent();
			this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
			this.SetStyle(ControlStyles.OptimizedDoubleBuffer, true);
			this.SetStyle(ControlStyles.UserPaint, true);
			this.SetStyle(ControlStyles.SupportsTransparentBackColor, true);
			this.SetStyle(ControlStyles.Opaque, true);
			this.TopMost = true;
			this.TintColor = Color.FromArgb(0x20, Color.Azure);
			Application.Idle += new EventHandler(Application_Idle);
		}

		void Application_Idle(object sender, EventArgs e) {
			//this.Invalidate();
		}

		protected override void OnPaint(PaintEventArgs e) {

			// If GlassBlur bitmap is null, we need to create a new one.
			if (this.GlassBlur == null) {

				this.GlassBlur = new Bitmap(this.ClientSize.Width + 2 * this.BlurFactor, this.ClientSize.Height + 2 * this.BlurFactor);
				this.GlassBlurGraphics = Graphics.FromImage(this.GlassBlur);
				this.GlassBlurGraphics.PixelOffsetMode = PixelOffsetMode.Half;

				this.GlassScratchArea = new RectangleF(0f, 0f, (float)this.ClientSize.Width / (float)this.BlurFactor, (float)this.ClientSize.Height / (float)this.BlurFactor);
				this.GlassBlurScratch = new Bitmap(
											(int)Math.Ceiling(GlassScratchArea.Width) + 2,
											(int)Math.Ceiling(GlassScratchArea.Height) + 2);

				this.GlassBlurScratchGraphics = Graphics.FromImage(this.GlassBlurScratch);
				this.GlassBlurScratchGraphics.InterpolationMode = InterpolationMode.HighQualityBicubic;
				this.GlassBlurScratchGraphics.PixelOffsetMode = PixelOffsetMode.HighQuality;

				this.CopyFromScreenSize = this.GlassBlur.Size;

				this.ScratchToScreenArea = new RectangleF(-this.BlurFactor, -this.BlurFactor, this.ClientSize.Width + this.BlurFactor * 2, this.ClientSize.Height + this.BlurFactor * 2);

				this.FinalClockFace = new Bitmap(this.ClientSize.Width, this.ClientSize.Height, PixelFormat.Format32bppArgb);
				this.FinalClockFaceGraphics = Graphics.FromImage(this.FinalClockFace);
				this.FinalClockFaceGraphics.PixelOffsetMode = PixelOffsetMode.Half;
				this.FinalClockFaceArea = this.ClientRectangle;

				this.AlphaWorkerArea = new int[this.ClientSize.Width * this.ClientSize.Height];

				// Get the face alpha:
				this.FaceAlpha = new int[this.AlphaWorkerArea.Length];
				using (Bitmap b = new Bitmap(this.FinalClockFace.Width, this.FinalClockFace.Height)) {
					using (Graphics g = Graphics.FromImage(b)) {
						g.PixelOffsetMode = PixelOffsetMode.HighQuality;
						g.SmoothingMode = SmoothingMode.HighQuality;
						using (Brush sb = new SolidBrush(Color.FromArgb(0xF0, Color.Black))) {
							g.FillEllipse(sb, 2, 2, this.FinalClockFace.Width - 3, this.FinalClockFace.Height - 3);
						}
						BitmapData fbd = b.LockBits(FinalClockFaceArea, ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
						Marshal.Copy(fbd.Scan0, this.FaceAlpha, 0, this.FaceAlpha.Length);
						b.UnlockBits(fbd);
					}
				}
			}

			// Copy the desktop area behind the window (and a bit extra!) to the GlassBlur area.
			this.GlassBlurGraphics.CopyFromScreen(this.Location.X - this.BlurFactor, this.Location.Y - this.BlurFactor, 0, 0, this.CopyFromScreenSize);

			// Draw it onto the "scratch" area:
			this.GlassBlurScratchGraphics.DrawImage(this.GlassBlur, this.GlassScratchArea);

			// Draw to the "final clock face".
			FinalClockFaceGraphics.DrawImage(this.GlassBlurScratch, this.ScratchToScreenArea, this.GlassScratchArea, GraphicsUnit.Pixel);

			BitmapData bd = FinalClockFace.LockBits(this.FinalClockFaceArea, ImageLockMode.ReadWrite, PixelFormat.Format32bppArgb);
			Marshal.Copy(bd.Scan0, this.AlphaWorkerArea, 0, this.AlphaWorkerArea.Length);


			int TransparencyMask = this.TransparencyKey.ToArgb();
			for (int i = 0; i < this.AlphaWorkerArea.Length; ++i) {

				this.AlphaWorkerArea[i] = unchecked((this.AlphaWorkerArea[i] & 0xFFFFFF) | (this.FaceAlpha[i] & (int)0xFF000000));

				//if ((this.AlphaWorkerArea[i] & 0xFF000000) == 0) this.AlphaWorkerArea[i] = TransparencyMask;
			}

			Marshal.Copy(this.AlphaWorkerArea, 0, bd.Scan0, this.AlphaWorkerArea.Length);
			FinalClockFace.UnlockBits(bd);


			e.Graphics.SmoothingMode = SmoothingMode.HighQuality;
			e.Graphics.DrawImageUnscaled(this.GlassBlur, -this.BlurFactor, -this.BlurFactor);
			e.Graphics.FillEllipse(Brushes.Azure, 2, 2, this.FinalClockFace.Width - 4, this.FinalClockFace.Height - 4);
			e.Graphics.DrawImageUnscaled(this.FinalClockFace, 0, 0);

			PointF Centre = new PointF(this.ClientSize.Width / 2f, this.ClientSize.Height / 2f);
			DateTime Now = DateTime.Now;

			PointF Seconds = new PointF(
								 (float)(+Math.Sin(Now.Second / 60f * Math.PI * 2f)),
								 (float)(-Math.Cos(Now.Second / 60f * Math.PI * 2f)));

			PointF Minutes = new PointF(
								 (float)(+Math.Sin((Now.Minute + Now.Second / 60f) / 60f * Math.PI * 2f)),
								 (float)(-Math.Cos((Now.Minute + Now.Second / 60f) / 60f * Math.PI * 2f)));

			PointF Hours = new PointF(
								 (float)(+Math.Sin((Now.Hour + Now.Minute / 60f) / 12f * Math.PI * 2f)),
								 (float)(-Math.Cos((Now.Hour + Now.Minute / 60f) / 12f * Math.PI * 2f)));



			e.Graphics.DrawLine(Pens.White,
				-.1f * Centre.X * Hours.X + Centre.X, -.1f * Centre.Y * Hours.Y + Centre.Y,
				+.5f * Centre.X * Hours.X + Centre.X, +.5f * Centre.Y * Hours.Y + Centre.Y);

			e.Graphics.DrawLine(Pens.White,
				-.2f * Centre.X * Minutes.X + Centre.X, -.2f * Centre.Y * Minutes.Y + Centre.Y,
				+.8f * Centre.X * Minutes.X + Centre.X, +.8f * Centre.Y * Minutes.Y + Centre.Y);

			e.Graphics.DrawLine(Pens.Red,
				-.3f * Centre.X * Seconds.X + Centre.X, -.3f * Centre.Y * Seconds.Y + Centre.Y,
				+.9f * Centre.X * Seconds.X + Centre.X, +.9f * Centre.Y * Seconds.Y + Centre.Y);

		}

		private void ClockUpdate_Tick(object sender, EventArgs e) {
			this.Invalidate();
		}

		private MouseEventArgs OriginalMouseDown;
		private void Clock_MouseDown(object sender, MouseEventArgs e) {
			if ((e.Button & MouseButtons.Left) == MouseButtons.Left) OriginalMouseDown = e;
		}

		private void Clock_MouseMove(object sender, MouseEventArgs e) {
			if (this.OriginalMouseDown != null && (e.Button & MouseButtons.Left) == MouseButtons.Left) {
				Point p = this.Location;
				p.Offset(e.X - OriginalMouseDown.X, e.Y - OriginalMouseDown.Y);
				this.Location = p;
				this.Refresh();
			}
		}

		private void Clock_DoubleClick(object sender, EventArgs e) {
			this.Close();
		}


	}
}
